@@ -1,7 +1,7 @@
 /*
  * OCaml bindings for alsa
  *
- * Copyright 2005-2006 Savonet team
+ * Copyright 2005-2020 Savonet team
  *
  * This file is part of ocaml-alsa.
  *
@@ -22,12 +22,11 @@
  * @author Samuel Mimram
  */
 
-/* $Id$ */
-
 #include <string.h>
 #include <alsa/asoundlib.h>
 
 #include <caml/alloc.h>
+#include <caml/bigarray.h>
 #include <caml/callback.h>
 #include <caml/custom.h>
 #include <caml/fail.h>
@@ -35,6 +34,7 @@
 #include <caml/misc.h>
 #include <caml/mlvalues.h>
 #include <caml/signals.h>
+#include <caml/threads.h>
 
 #include <assert.h>
 
@@ -49,6 +49,8 @@ typedef struct my_snd_pcm_t
   int frame_size;
 } my_snd_pcm_t;
 
+/********** PCM **********/
+
 static struct custom_operations pcm_handle_ops =
 {
   "ocaml_alsa_pcm_handle",
@@ -115,11 +117,7 @@ static int int_of_pcm_mode(value mode)
   return ans;
 }
 
-/* Here, we use the type snd_pcm_sframes_t
- * which is long usually and makes sure we
- * avoid long to int overflow in the read/write
- * function's return value. */
-static void check_for_err(snd_pcm_sframes_t ret)
+static void check_for_err(int ret)
 {
   if (ret >= 0)
     return;
@@ -162,6 +160,10 @@ static void check_for_err(snd_pcm_sframes_t ret)
       caml_raise_constant(*caml_named_value("alsa_exn_device_busy"));
       break;
 
+    case EAGAIN:
+      caml_raise_constant(*caml_named_value("alsa_exn_try_again"));
+      break;
+
     default:
       caml_raise_with_arg(*caml_named_value("alsa_exn_unknown_error"), Val_int(-ret));
       break;
@@ -171,7 +173,7 @@ static void check_for_err(snd_pcm_sframes_t ret)
 CAMLprim value ocaml_snd_int_of_error(value name)
 {
   CAMLparam1(name);
-  char *s = String_val(name);
+  const char *s = String_val(name);
   if (!strcmp(s,"alsa_exn_io_error"))
     CAMLreturn(Val_int(-EIO));
   if (!strcmp(s,"alsa_exn_device_busy"))
@@ -228,7 +230,7 @@ static struct custom_operations hw_params_ops =
 
 static value create_hw_params()
 {
-  CAMLlocal1(ans);
+  value ans;
   snd_pcm_hw_params_t *params;
 
   check_for_err(snd_pcm_hw_params_malloc(&params));
@@ -311,7 +313,7 @@ CAMLprim value ocaml_snd_pcm_readi(value handle_, value dbuf, value ofs_, value
   ret = snd_pcm_readi(handle, buf, len);
   caml_leave_blocking_section();
 
-  memcpy(String_val(dbuf) + ofs, buf, len * Frame_size_val(handle_));
+  memcpy((void*)String_val(dbuf) + ofs, buf, len * Frame_size_val(handle_));
   free(buf);
   check_for_err(ret);
 
@@ -364,7 +366,7 @@ CAMLprim value ocaml_snd_pcm_readn(value handle_, value dbuf, value ofs_, value
 
   for(c = 0; c < chans; c++)
   {
-    memcpy(String_val(Field(dbuf, c)) + ofs, buf[c], 2 * len);
+    memcpy((void*)String_val(Field(dbuf, c)) + ofs, buf[c], 2 * len);
     free(buf[c]);
   }
   free(buf);
@@ -468,6 +470,94 @@ CAMLprim value ocaml_snd_pcm_writen_float(value handle_, value fbuf, value ofs_,
   CAMLreturn(Val_int(ret));
 }
 
+CAMLprim value ocaml_snd_pcm_readn_float_ba(value handle_, value dbuf)
+{
+  CAMLparam2(handle_, dbuf);
+  int chans = Wosize_val(dbuf);
+  snd_pcm_t *handle = Pcm_handle_val(handle_);
+  int len = 0;
+  float **buf;
+  struct caml_ba_array *ba;
+  snd_pcm_sframes_t ret;
+  int c;
+
+  buf = malloc(chans * sizeof(float*));
+  for(c = 0; c < chans; c++)
+    {
+      ba = Caml_ba_array_val(Field(dbuf, c));
+      if (c == 0)
+        len = ba->dim[0];
+      else
+        {
+          if (ba->dim[0] != len)
+            caml_failwith("Invalid argument");
+        }
+      buf[c] = ba->data;
+    }
+
+  caml_enter_blocking_section();
+  ret = snd_pcm_readn(handle, (void**)buf, len);
+  caml_leave_blocking_section();
+
+  free(buf);
+  check_for_err(ret);
+
+  CAMLreturn(Val_int(ret));
+}
+
+CAMLprim value ocaml_snd_pcm_writen_float_ba(value handle_, value fbuf)
+{
+  CAMLparam2(handle_, fbuf);
+  int chans = Wosize_val(fbuf);
+  snd_pcm_t *handle = Pcm_handle_val(handle_);
+  int len = 0;
+  float **buf = malloc(chans * sizeof(float*));
+  snd_pcm_sframes_t ret;
+  struct caml_ba_array *ba;
+  int c;
+
+  for(c = 0; c < chans; c++)
+  {
+    ba = Caml_ba_array_val(Field(fbuf, c));
+    if (c == 0)
+      len = ba->dim[0];
+    else
+      {
+        if (ba->dim[0] != len)
+          caml_failwith("Invalid argument");
+      }
+    buf[c] = ba->data;
+  }
+
+  caml_enter_blocking_section();
+  ret = snd_pcm_writen(handle, (void**)buf, len);
+  caml_leave_blocking_section();
+
+  free(buf);
+  check_for_err(ret);
+
+  CAMLreturn(Val_int(ret));
+}
+
+CAMLprim value ocaml_snd_pcm_writei_float_ba(value handle_, value channels, value fbuf)
+{
+  CAMLparam3(handle_, channels, fbuf);
+  snd_pcm_t *handle = Pcm_handle_val(handle_);
+  int chans = Int_val(channels);
+  struct caml_ba_array *ba = Caml_ba_array_val(fbuf);
+  int len = ba->dim[0];
+  float *buf = ba->data;
+  snd_pcm_sframes_t ret;
+
+  caml_enter_blocking_section();
+  ret = snd_pcm_writei(handle, buf, len / chans);
+  caml_leave_blocking_section();
+
+  check_for_err(ret);
+
+  CAMLreturn(Val_int(ret));
+}
+
 CAMLprim value ocaml_snd_pcm_readn_float64(value handle_, value dbuf, value ofs_, value len_)
 {
   CAMLparam4(handle_, dbuf, ofs_, len_);
@@ -768,6 +858,16 @@ CAMLprim value ocaml_snd_pcm_get_periods_max(value params)
   CAMLreturn(result);
 }
 
+CAMLprim value ocaml_snd_pcm_get_period_size(value params)
+{
+  CAMLparam1(params);
+  snd_pcm_uframes_t ans;
+
+  check_for_err(snd_pcm_hw_params_get_period_size(Hw_params_val(params), &ans, 0));
+
+  CAMLreturn(Val_int(ans));
+}
+
 CAMLprim value ocaml_snd_pcm_set_buffer_size(value handle, value params, value size)
 {
   CAMLparam3(handle, params, size);
@@ -787,6 +887,16 @@ CAMLprim value ocaml_snd_pcm_set_buffer_size_near(value handle, value params, va
   CAMLreturn(Val_int(s));
 }
 
+CAMLprim value ocaml_snd_pcm_get_buffer_size(value params)
+{
+  CAMLparam1(params);
+  snd_pcm_uframes_t ans;
+
+  check_for_err(snd_pcm_hw_params_get_buffer_size(Hw_params_val(params), &ans));
+
+  CAMLreturn(Val_int(ans));
+}
+
 CAMLprim value ocaml_snd_pcm_get_buffer_size_min(value params)
 {
   CAMLparam1(params);
@@ -837,3 +947,278 @@ CAMLprim value ocaml_snd_pcm_set_nonblock(value handle, value nonblocking)
 
   CAMLreturn(Val_unit);
 }
+
+/********** Sequencer **********/
+
+static struct custom_operations seq_handle_ops =
+{
+  "ocaml_alsa_seq_handle",
+  NULL,
+  custom_compare_default,
+  custom_hash_default,
+  custom_serialize_default,
+  custom_deserialize_default
+};
+
+#define Seq_val(v) (*((snd_seq_t**)Data_custom_val(v)))
+
+CAMLprim value ocaml_snd_seq_open(value name, value stream_, value mode_)
+{
+  CAMLparam3(name, stream_, mode_);
+  CAMLlocal1(seq);
+
+  int stream = Int_val(stream_);
+  int mode = Int_val(mode_);
+
+  snd_seq_t *seq_handle = NULL;
+  int ret;
+
+  seq = caml_alloc_custom(&seq_handle_ops, sizeof(snd_seq_t*), 0, 1);
+
+  ret = snd_seq_open(&seq_handle, String_val(name), stream, mode);
+  check_for_err(ret);
+
+  Seq_val(seq) = seq_handle;
+
+  CAMLreturn(seq);
+}
+
+CAMLprim value ocaml_snd_seq_set_client_name(value seq, value name)
+{
+  CAMLparam2(seq, name);
+  int ret;
+
+  ret = snd_seq_set_client_name(Seq_val(seq), String_val(name));
+  check_for_err(ret);
+
+  CAMLreturn(Val_unit);
+}
+
+CAMLprim value ocaml_snd_seq_create_port(value seq, value name, value _caps, value _type)
+{
+  CAMLparam4(seq, name, _caps, _type);
+  unsigned int caps = 0;
+  unsigned int type = 0;
+  static unsigned int capsv[] = {SND_SEQ_PORT_CAP_READ, SND_SEQ_PORT_CAP_WRITE, SND_SEQ_PORT_CAP_SYNC_READ, SND_SEQ_PORT_CAP_SYNC_WRITE, SND_SEQ_PORT_CAP_DUPLEX, SND_SEQ_PORT_CAP_SUBS_READ, SND_SEQ_PORT_CAP_SUBS_WRITE, SND_SEQ_PORT_CAP_NO_EXPORT};
+  static unsigned int typev[] = {SND_SEQ_PORT_TYPE_SPECIFIC, SND_SEQ_PORT_TYPE_MIDI_GENERIC, SND_SEQ_PORT_TYPE_MIDI_GM, SND_SEQ_PORT_TYPE_MIDI_GM2, SND_SEQ_PORT_TYPE_MIDI_GS, SND_SEQ_PORT_TYPE_MIDI_XG, SND_SEQ_PORT_TYPE_MIDI_MT32, SND_SEQ_PORT_TYPE_HARDWARE, SND_SEQ_PORT_TYPE_SOFTWARE, SND_SEQ_PORT_TYPE_SYNTHESIZER, SND_SEQ_PORT_TYPE_PORT, SND_SEQ_PORT_TYPE_APPLICATION};
+  int port;
+
+  while (_caps != Val_emptylist)
+    {
+      caps |= capsv[Int_val(Field(_caps, 0))];
+      _caps = Field(_caps, 1);
+    }
+  while (_type != Val_emptylist)
+    {
+      type |= typev[Int_val(Field(_type, 0))];
+      _type = Field(_type, 1);
+    }
+
+  port = snd_seq_create_simple_port(Seq_val(seq), String_val(name), caps, type);
+  check_for_err(port);
+
+  CAMLreturn(Val_int(port));
+}
+
+/* Read from every possible port */
+CAMLprim value ocaml_snd_subscribe_read_all(value _seq, value _dst)
+{
+  CAMLparam2(_seq, _dst);
+  snd_seq_t *seq = Seq_val(_seq);
+  int dst = Int_val(dst);
+  snd_seq_client_info_t *cinfo;
+  snd_seq_port_info_t *pinfo;
+  snd_seq_client_info_alloca(&cinfo);
+  snd_seq_port_info_alloca(&pinfo);
+  snd_seq_client_info_set_client(cinfo, -1);
+  while (snd_seq_query_next_client(seq, cinfo) >= 0)
+    {
+      snd_seq_port_info_set_client(pinfo, snd_seq_client_info_get_client(cinfo));
+      snd_seq_port_info_set_port(pinfo, -1);
+      // Avoid system device, midi through and self
+      while (snd_seq_client_info_get_client(cinfo) != 0 && snd_seq_client_info_get_client(cinfo) != 14 && snd_seq_client_info_get_client(cinfo) != snd_seq_client_id(seq) && snd_seq_query_next_port(seq, pinfo) >= 0)
+        {
+          if ((snd_seq_port_info_get_capability(pinfo) & (SND_SEQ_PORT_CAP_READ|SND_SEQ_PORT_CAP_SUBS_READ)) == (SND_SEQ_PORT_CAP_READ|SND_SEQ_PORT_CAP_SUBS_READ))
+            {
+              caml_release_runtime_system();
+              snd_seq_addr_t sender, dest;
+              sender.client = snd_seq_client_info_get_client(cinfo);
+              sender.port = snd_seq_port_info_get_port(pinfo);
+              dest.client = snd_seq_client_id(seq);
+              dest.port = dst;
+              snd_seq_port_subscribe_t *subs;
+              snd_seq_port_subscribe_alloca(&subs);
+              snd_seq_port_subscribe_set_sender(subs, &sender);
+              snd_seq_port_subscribe_set_dest(subs, &dest);
+              snd_seq_subscribe_port(seq, subs);
+              caml_acquire_runtime_system();
+            }
+        }
+    }
+  CAMLreturn(Val_unit);
+}
+
+CAMLprim value ocaml_snd_subscribe_write_all(value _seq, value _src)
+{
+  CAMLparam2(_seq, _src);
+  snd_seq_t *seq = Seq_val(_seq);
+  int src = Int_val(src);
+  snd_seq_client_info_t *cinfo;
+  snd_seq_port_info_t *pinfo;
+  snd_seq_client_info_alloca(&cinfo);
+  snd_seq_port_info_alloca(&pinfo);
+  snd_seq_client_info_set_client(cinfo, -1);
+  while (snd_seq_query_next_client(seq, cinfo) >= 0)
+    {
+      snd_seq_port_info_set_client(pinfo, snd_seq_client_info_get_client(cinfo));
+      snd_seq_port_info_set_port(pinfo, -1);
+      while (snd_seq_client_info_get_client(cinfo) != 0 && snd_seq_client_info_get_client(cinfo) != 14 && snd_seq_client_info_get_client(cinfo) != snd_seq_client_id(seq) && snd_seq_query_next_port(seq, pinfo) >= 0)
+        {
+          if ((snd_seq_port_info_get_capability(pinfo) & (SND_SEQ_PORT_CAP_WRITE|SND_SEQ_PORT_CAP_SUBS_WRITE)) == (SND_SEQ_PORT_CAP_WRITE|SND_SEQ_PORT_CAP_SUBS_WRITE))
+            {
+              caml_release_runtime_system();
+              snd_seq_addr_t sender, dest;
+              sender.client = snd_seq_client_id(seq);
+              sender.port = src;
+              dest.client = snd_seq_client_info_get_client(cinfo);
+              dest.port = snd_seq_port_info_get_port(pinfo);
+              snd_seq_port_subscribe_t *subs;
+              snd_seq_port_subscribe_alloca(&subs);
+              snd_seq_port_subscribe_set_sender(subs, &sender);
+              snd_seq_port_subscribe_set_dest(subs, &dest);
+              snd_seq_subscribe_port(seq, subs);
+              caml_acquire_runtime_system();
+            }
+        }
+    }
+  CAMLreturn(Val_unit);
+}
+
+static value Val_note(snd_seq_ev_note_t n)
+{
+  CAMLparam0();
+  CAMLlocal1(ans);
+  ans = caml_alloc_tuple(5);
+  Store_field(ans, 0, Val_int(n.channel));
+  Store_field(ans, 1, Val_int(n.note));
+  Store_field(ans, 2, Val_int(n.velocity));
+  Store_field(ans, 3, Val_int(n.off_velocity));
+  Store_field(ans, 4, Val_int(n.duration));
+  CAMLreturn(ans);
+}
+
+static value Val_controller(snd_seq_ev_ctrl_t c)
+{
+  CAMLparam0();
+  CAMLlocal1(ans);
+  ans = caml_alloc_tuple(3);
+  Store_field(ans, 0, Val_int(c.channel));
+  Store_field(ans, 1, Val_int(c.param));
+  Store_field(ans, 2, Val_int(c.value));
+  CAMLreturn(ans);
+}
+
+CAMLprim value ocaml_snd_seq_event_input(value handle)
+{
+  CAMLparam1(handle);
+  CAMLlocal1(ans);
+  CAMLlocal1(event);
+
+  snd_seq_t *seq_handle = Seq_val(handle);
+  snd_seq_event_t *ev = NULL;
+  int ret = 0;
+
+  caml_release_runtime_system();
+  ret = snd_seq_event_input(seq_handle, &ev);
+  caml_acquire_runtime_system();
+
+  check_for_err(ret);
+  assert(ev);
+
+  switch (ev->type)
+    {
+    case SND_SEQ_EVENT_NOTEON:
+      event = caml_alloc(1, 3);
+      Store_field(event, 0, Val_note(ev->data.note));
+      break;
+
+    case SND_SEQ_EVENT_NOTEOFF:
+      event = caml_alloc(1, 4);
+      Store_field(event, 0, Val_note(ev->data.note));
+      break;
+
+    case SND_SEQ_EVENT_CONTROLLER:
+      event = caml_alloc(1, 6);
+      Store_field(event, 0, Val_controller(ev->data.control));
+      break;
+
+    case SND_SEQ_EVENT_PGMCHANGE:
+      event = caml_alloc(1, 7);
+      Store_field(event, 0, Val_controller(ev->data.control));
+      break;
+
+    case SND_SEQ_EVENT_PITCHBEND:
+      event = caml_alloc(1, 9);
+      Store_field(event, 0, Val_controller(ev->data.control));
+      break;
+
+    default:
+      // TODO: change this number when adding new constructors...
+      event = caml_alloc(1, 10);
+      Store_field(event, 0, Val_int(ev->type));
+      break;
+    }
+
+  ans = caml_alloc_tuple(2);
+  Store_field(ans, 0, event);
+  Store_field(ans, 1, Val_unit);
+
+  CAMLreturn(ans);
+}
+
+CAMLprim value ocaml_snd_seq_event_output(value handle, value event)
+{
+  CAMLparam2(handle, event);
+  CAMLlocal1(v);
+
+  snd_seq_t *seq_handle = Seq_val(handle);
+  snd_seq_event_t ev;
+  int ret = 0;
+
+  snd_seq_ev_clear(&ev);
+  //snd_seq_ev_set_source(&ev, Int_val(port));
+  snd_seq_ev_set_subs(&ev);
+  snd_seq_ev_set_direct(&ev);
+
+  switch (Tag_val(event))
+    {
+    case 3:
+      v = Field(event, 0);
+      snd_seq_ev_set_noteon(&ev, Int_val(Field(v, 0)), Int_val(Field(v, 1)), Int_val(Field(v, 2)));
+      break;
+
+    case 4:
+      v = Field(event, 0);
+      snd_seq_ev_set_noteoff(&ev, Int_val(Field(v, 0)), Int_val(Field(v, 1)), Int_val(Field(v, 2)));
+      break;
+
+    default:
+      caml_failwith("TODO: unhandled constructor");
+      break;
+    }
+
+
+  caml_release_runtime_system();
+  ret = snd_seq_event_output(seq_handle, &ev);
+  caml_acquire_runtime_system();
+  check_for_err(ret);
+
+  // TODO: other function for this
+  caml_release_runtime_system();
+  ret = snd_seq_drain_output(seq_handle);
+  caml_acquire_runtime_system();
+  check_for_err(ret);
+
+  CAMLreturn(Val_unit);
+}
+
